@@ -25,6 +25,14 @@
 /*	MAC_EXPAND_RELOP_FN relop_eval)
 /*
 /*	MAC_EXP_OP_RES mac_exp_op_res_bool[2];
+/
+/*	typedef int (*MAC_EXPAND_NAMED_FN) (
+/*	VSTRING *out,
+/*	const char *arg)
+/*
+/*	void	mac_expand_add_named_fn(
+/*	const char *name,
+/*	MAC_EXPAND_NAMED_FN action)
 /* DESCRIPTION
 /*	This module implements parameter-less named attribute
 /*	expansions, both conditional and unconditional. As of Postfix
@@ -51,6 +59,11 @@
 /*	named attribute expansion and relational expression evaluation.
 /*	Otherwise, the result is empty.  Whitespace before or after
 /*	{text} is ignored.
+/* .IP "${name{text}}"
+/*	Apply the registered function \fIname\fR to the specified text
+/*	after named attribute expansion and expression evaluation,
+/*	and replace this input with the function result. Functions are
+/*	registered with mac_expand_add_named_fn().
 /* .IP "${name:text}, ${name:{text}}"
 /*	Conditional attribute-based substitution. If the attribute
 /*	value is empty or undefined, the expansion is the given
@@ -133,11 +146,18 @@
 /*	mac_exp_op_res_bool provides an array that converts a boolean
 /*	value (0 or 1) to the corresponding MAX_EXP_OP_RES_TRUE or
 /*	MAX_EXP_OP_RES_FALSE value.
+/*
+/*	mac_expand_add_named_fn() registers a C function that may be
+/*	called as ${name{text}}. The function input is the text after
+/*	attribute expansion and expression evaluation. The function
+/*	should append its output to the specified buffer, and it
+/*	should return either MAC_PARSE_OK or MAC_PARSE_ERROR.
 /* DIAGNOSTICS
 /*	Fatal errors: out of memory.  Warnings: syntax errors, unreasonable
 /*	recursion depth.
 /*
-/*	The result value is the binary OR of zero or more of the following:
+/*	The mac_expand() result value is the binary OR of zero or more
+/*	of the following:
 /* .IP MAC_PARSE_ERROR
 /*	A syntax error was found in \fBpattern\fR, or some attribute had
 /*	an unreasonable nesting depth.
@@ -159,6 +179,9 @@
 /*	Google, Inc.
 /*	111 8th Avenue
 /*	New York, NY 10011, USA
+/*
+/*	Wietse Venema
+/*	porcupine.org
 /*--*/
 
 /* System library. */
@@ -263,6 +286,14 @@ static const NAME_CODE mac_exp_op_table[] =
 static HTABLE *mac_exp_ext_table;
 static VSTRING *mac_exp_ext_key;
 
+ /*
+  * Support for named functions.
+  */
+static HTABLE *mac_exp_named_fn_table;
+struct mac_exp_named_fn_entry {
+    MAC_EXPAND_NAMED_FN action;
+};
+
  /*
   * SLMs.
   */
@@ -360,18 +391,27 @@ static int PRINTFLIKE(2, 3) mac_exp_parse_error(MAC_EXP_CONTEXT *mc,
 
 /* mac_exp_extract_curly_payload - balance {}, skip whitespace, return payload */
 
-static char *mac_exp_extract_curly_payload(MAC_EXP_CONTEXT *mc, char **bp)
+static char *mac_exp_extract_curly_payload(MAC_EXP_CONTEXT *mc, char **bp,
+					           int strip_space)
 {
     char   *payload;
     char   *cp;
     int     level;
     int     ch;
 
+#define DO_STRIP_SPACE	1
+#define NO_STRIP_SPACE	0
+
     /*
      * Extract the payload and balance the {}. The caller is expected to skip
      * leading whitespace before the {. See MAC_EXP_FIND_LEFT_CURLY().
+     * TODO(wietse) this code pre-dates extpar() and mainly differs in how an
+     * error is reported.
      */
-    for (level = 1, cp = *bp, payload = ++cp; /* see below */ ; cp++) {
+    payload = *bp + 1;
+    if (strip_space)
+	payload += strspn(payload, MAC_EXP_WHITESPACE);
+    for (level = 1, cp = payload; /* see below */ ; cp++) {
 	if ((ch = *cp) == 0) {
 	    mac_exp_parse_error(mc, "unbalanced {} in attribute expression: "
 				"\"%s\"",
@@ -384,6 +424,8 @@ static char *mac_exp_extract_curly_payload(MAC_EXP_CONTEXT *mc, char **bp)
 		break;
 	}
     }
+    if (strip_space)
+	trimblanks(payload, cp - payload)[0] = 0;
     *cp++ = 0;
 
     /*
@@ -417,7 +459,8 @@ static int mac_exp_parse_relational(MAC_EXP_CONTEXT *mc, const char **lookup,
      * Left operand. The caller is expected to skip leading whitespace before
      * the {. See MAC_EXP_FIND_LEFT_CURLY().
      */
-    if ((left_op_strval = mac_exp_extract_curly_payload(mc, &cp)) == 0)
+    if ((left_op_strval = mac_exp_extract_curly_payload(mc, &cp,
+						      NO_STRIP_SPACE)) == 0)
 	return (mc->status);
 
     /*
@@ -458,7 +501,8 @@ static int mac_exp_parse_relational(MAC_EXP_CONTEXT *mc, const char **lookup,
 	MAC_EXP_ERR_RETURN(mc, "\"{expression}\" expected at: "
 			   "\"...{%s} %.*s>>>%.20s\"",
 			   left_op_strval, (int) op_len, op_pos, cp);
-    if ((rite_op_strval = mac_exp_extract_curly_payload(mc, &cp)) == 0)
+    if ((rite_op_strval = mac_exp_extract_curly_payload(mc, &cp,
+						      NO_STRIP_SPACE)) == 0)
 	return (mc->status);
 
     /*
@@ -535,6 +579,70 @@ void    mac_expand_add_relop(int *tok_list, const char *suffix,
     }
 }
 
+/* mac_expand_add_named_fn - register named-function callback */
+
+void    mac_expand_add_named_fn(const char *name, MAC_EXPAND_NAMED_FN action)
+{
+    struct mac_exp_named_fn_entry *xp;
+
+    /*
+     * Sanity checks.
+     */
+    if (!allalnumus(name))
+	msg_panic("%s: bad function name: \"%s\"", __func__, name);
+
+    /*
+     * One-time initialization.
+     */
+    if (mac_exp_named_fn_table == 0)
+	mac_exp_named_fn_table = htable_create(10);
+
+    /*
+     * The C language spec allows sizeof(void *) < sizeof(function pointer).
+     */
+    if (htable_locate(mac_exp_named_fn_table, name) != 0)
+	msg_panic("%s: duplicate key: %s", __func__, name);
+    xp = (struct mac_exp_named_fn_entry *) mymalloc(sizeof(*xp));
+    xp->action = action;
+    (void) htable_enter(mac_exp_named_fn_table, name, (void *) xp);
+}
+
+/* mac_exp_parse_function - interpolate result from caller-defined function */
+
+static int mac_exp_parse_function(char *name_start, char *name_end,
+				          char *cp, MAC_EXP_CONTEXT *mc)
+{
+    char   *fn_arg;
+    struct mac_exp_named_fn_entry *xp;
+    VSTRING *buf = 0;
+
+    /* cp is positioned at the '{', zero or more spaces after name_end. */
+    if ((fn_arg = mac_exp_extract_curly_payload(mc, &cp, DO_STRIP_SPACE)) == 0)
+	return (mc->status);
+    if (*cp != 0)				/* garbage */
+	MAC_EXP_ERR_RETURN(mc, "unexpected input at: "
+			   "\"...%s}>>>%.20s\"", fn_arg, cp);
+    *name_end = 0;
+    /* Look up the function and evaluate. */
+    xp = (struct mac_exp_named_fn_entry *)
+	htable_find(mac_exp_named_fn_table, name_start);
+    if (xp == 0)
+	MAC_EXP_ERR_RETURN(mc, "unknown function \"%s\" at"
+			"\"...>>>%s{%s}\"", name_start, name_start, fn_arg);
+    if ((mc->flags & MAC_EXP_FLAG_SCAN) == 0)
+	buf = vstring_alloc(100);
+    mc->status |=
+	mac_expand(buf, fn_arg, mc->flags, mc->filter, mc->lookup,
+		   mc->context);
+    if ((mc->flags & MAC_EXP_FLAG_SCAN) == 0) {
+	if ((mc->status & MAC_PARSE_ERROR) == 0
+	    && xp->action(mc->result, vstring_str(buf)) == MAC_PARSE_ERROR)
+	    mc->status |= MAC_PARSE_ERROR;
+	vstring_free(buf);
+    }
+    return (mc->status);
+}
+
 /* mac_expand_callback - callback for mac_parse */
 
 static int mac_expand_callback(int type, VSTRING *buf, void *ptr)
@@ -560,8 +668,9 @@ static int mac_expand_callback(int type, VSTRING *buf, void *ptr)
 	return (mc->status);
 
     /*
-     * Named parameter or relational expression. In case of a syntax error,
-     * return without doing damage, and issue a warning instead.
+     * Named parameter, function call, or relational expression. In case of a
+     * syntax error, return without doing damage, and issue a warning
+     * instead.
      */
     if (type == MAC_PARSE_EXPR) {
 
@@ -587,14 +696,16 @@ static int mac_expand_callback(int type, VSTRING *buf, void *ptr)
 	}
 
 	/*
-	 * Named parameter.
+	 * Named parameter or function call.
 	 */
 	else {
 	    char   *start;
 
 	    /*
-	     * Look for the ? or : operator. In case of a syntax error,
-	     * return without doing damage, and issue a warning instead.
+	     * Collect the name of the parameter or function. Look for the ?
+	     * or : operator, or the '{' to indicate a function call. In case
+	     * of a syntax error, return without doing damage, and issue a
+	     * warning instead.
 	     */
 	    start = (cp += strspn(cp, MAC_EXP_WHITESPACE));
 	    for ( /* void */ ; /* void */ ; cp++) {
@@ -609,6 +720,8 @@ static int mac_expand_callback(int type, VSTRING *buf, void *ptr)
 		    lookup_mode = MAC_EXP_MODE_TEST;
 		    break;
 		}
+		if (ch == '{')
+		    return (mac_exp_parse_function(start, cp, cp + tmp_len, mc));
 		ch = *cp;
 		if (!ISALNUM(ch) && ch != '_') {
 		    MAC_EXP_ERR_RETURN(mc, "attribute name syntax error at: "
@@ -638,7 +751,8 @@ static int mac_expand_callback(int type, VSTRING *buf, void *ptr)
 	switch (ch) {
 	case '?':
 	    if (MAC_EXP_FIND_LEFT_CURLY(tmp_len, cp)) {
-		if ((res_iftrue = mac_exp_extract_curly_payload(mc, &cp)) == 0)
+		if ((res_iftrue = mac_exp_extract_curly_payload(mc, &cp,
+						      NO_STRIP_SPACE)) == 0)
 		    return (mc->status);
 	    } else {
 		res_iftrue = cp;
@@ -656,7 +770,8 @@ static int mac_expand_callback(int type, VSTRING *buf, void *ptr)
 	    /* FALLTHROUGH: do not remove, see comment above. */
 	case ':':
 	    if (MAC_EXP_FIND_LEFT_CURLY(tmp_len, cp)) {
-		if ((res_iffalse = mac_exp_extract_curly_payload(mc, &cp)) == 0)
+		if ((res_iffalse = mac_exp_extract_curly_payload(mc, &cp,
+						      NO_STRIP_SPACE)) == 0)
 		    return (mc->status);
 	    } else {
 		res_iffalse = cp;
@@ -776,6 +891,12 @@ static MAC_EXP_OP_RES length_relop_eval(const char *left, int relop,
     }
 }
 
+static int named_fn_eval(VSTRING *out, const char *in)
+{
+    vstring_sprintf_append(out, "-=oO%sOo=-", in);
+    return (0);
+}
+
 int     main(int unused_argc, char **argv)
 {
     VSTRING *buf = vstring_alloc(100);
@@ -797,6 +918,11 @@ int     main(int unused_argc, char **argv)
      */
     mac_expand_add_relop(length_relops, "length", length_relop_eval);
 
+    /*
+     * Add a silly named function that decorates its argument.
+     */
+    mac_expand_add_named_fn("test_named_fn", named_fn_eval);
+
     /*
      * Loop over the inputs.
      */
